@@ -3,32 +3,36 @@ import sys
 
 import numpy as np
 import scipy.sparse as sp
+import joblib
 
 from io import StringIO
 
 from sklearn.base import BaseEstimator, ClassifierMixin
 from sklearn.utils import deprecated
-from sklearn.utils import _joblib
-from sklearn.utils.testing import (assert_raises_regex,
-                                   assert_equal, ignore_warnings,
-                                   assert_warns, assert_raises)
-from sklearn.utils.estimator_checks import check_estimator
+from sklearn.utils._testing import (assert_raises_regex,
+                                   ignore_warnings,
+                                   assert_warns, assert_raises,
+                                   SkipTest)
+from sklearn.utils.estimator_checks import check_estimator, _NotAnArray
 from sklearn.utils.estimator_checks \
     import check_class_weight_balanced_linear_classifier
 from sklearn.utils.estimator_checks import set_random_state
-from sklearn.utils.estimator_checks import set_checking_parameters
+from sklearn.utils.estimator_checks import _set_checking_parameters
 from sklearn.utils.estimator_checks import check_estimators_unfitted
 from sklearn.utils.estimator_checks import check_fit_score_takes_y
 from sklearn.utils.estimator_checks import check_no_attributes_set_in_init
+from sklearn.utils.validation import check_is_fitted
 from sklearn.utils.estimator_checks import check_outlier_corruption
-from sklearn.ensemble import AdaBoostClassifier, RandomForestClassifier
+from sklearn.utils.fixes import _parse_version
+from sklearn.ensemble import RandomForestClassifier
 from sklearn.linear_model import LinearRegression, SGDClassifier
 from sklearn.mixture import GaussianMixture
 from sklearn.cluster import MiniBatchKMeans
 from sklearn.decomposition import NMF
-from sklearn.linear_model import MultiTaskElasticNet
+from sklearn.linear_model import MultiTaskElasticNet, LogisticRegression
 from sklearn.svm import SVC
 from sklearn.neighbors import KNeighborsRegressor
+from sklearn.tree import DecisionTreeClassifier
 from sklearn.utils.validation import check_X_y, check_array
 
 
@@ -40,7 +44,7 @@ class CorrectNotFittedError(ValueError):
     """
 
 
-class BaseBadClassifier(BaseEstimator, ClassifierMixin):
+class BaseBadClassifier(ClassifierMixin, BaseEstimator):
     def fit(self, X, y):
         return self
 
@@ -167,8 +171,7 @@ class CorrectNotFittedErrorClassifier(BaseBadClassifier):
         return self
 
     def predict(self, X):
-        if not hasattr(self, 'coef_'):
-            raise CorrectNotFittedError("estimator is not fitted yet")
+        check_is_fitted(self)
         X = check_array(X)
         return np.ones(X.shape[0])
 
@@ -276,6 +279,44 @@ class SparseTransformer(BaseEstimator):
         return sp.csr_matrix(X)
 
 
+class UntaggedBinaryClassifier(DecisionTreeClassifier):
+    # Toy classifier that only supports binary classification, will fail tests.
+    def fit(self, X, y, sample_weight=None):
+        super().fit(X, y, sample_weight)
+        if np.all(self.n_classes_ > 2):
+            raise ValueError('Only 2 classes are supported')
+        return self
+
+
+class TaggedBinaryClassifier(UntaggedBinaryClassifier):
+    # Toy classifier that only supports binary classification.
+    def _more_tags(self):
+        return {'binary_only': True}
+
+
+class RequiresPositiveYRegressor(LinearRegression):
+
+    def fit(self, X, y):
+        X, y = check_X_y(X, y, multi_output=True)
+        if (y <= 0).any():
+            raise ValueError('negative y values not supported!')
+        return super().fit(X, y)
+
+    def _more_tags(self):
+        return {"requires_positive_y": True}
+
+
+def test_not_an_array_array_function():
+    np_version = _parse_version(np.__version__)
+    if np_version < (1, 17):
+        raise SkipTest("array_function protocol not supported in numpy <1.17")
+    not_array = _NotAnArray(np.ones(10))
+    msg = "Don't want to call array_function sum!"
+    assert_raises_regex(TypeError, msg, np.sum, not_array)
+    # always returns True
+    assert np.may_share_memory(not_array, None)
+
+
 def test_check_fit_score_takes_y_works_on_deprecated_fit():
     # Tests that check_fit_score_takes_y works on a class with
     # a deprecated fit method
@@ -376,15 +417,28 @@ def test_check_estimator():
     assert_raises_regex(AssertionError, msg, check_estimator,
                         LargeSparseNotSupportedClassifier)
 
+    # does error on binary_only untagged estimator
+    msg = 'Only 2 classes are supported'
+    assert_raises_regex(ValueError, msg, check_estimator,
+                        UntaggedBinaryClassifier)
+
     # non-regression test for estimators transforming to sparse data
     check_estimator(SparseTransformer())
 
     # doesn't error on actual estimator
-    check_estimator(AdaBoostClassifier)
-    check_estimator(AdaBoostClassifier())
+    check_estimator(LogisticRegression)
+    check_estimator(LogisticRegression(C=0.01))
     check_estimator(MultiTaskElasticNet)
     check_estimator(MultiTaskElasticNet())
 
+    # doesn't error on binary_only tagged estimator
+    check_estimator(TaggedBinaryClassifier)
+
+    # Check regressor with requires_positive_y estimator tag
+    msg = 'negative y values not supported!'
+    assert_raises_regex(ValueError, msg, check_estimator,
+                        RequiresPositiveYRegressor)
+
 
 def test_check_outlier_corruption():
     # should raise AssertionError
@@ -409,32 +463,32 @@ def test_check_estimator_clones():
     for Estimator in [GaussianMixture, LinearRegression,
                       RandomForestClassifier, NMF, SGDClassifier,
                       MiniBatchKMeans]:
-        with ignore_warnings(category=(FutureWarning, DeprecationWarning)):
+        with ignore_warnings(category=FutureWarning):
             # when 'est = SGDClassifier()'
             est = Estimator()
-            set_checking_parameters(est)
+            _set_checking_parameters(est)
             set_random_state(est)
             # without fitting
-            old_hash = _joblib.hash(est)
+            old_hash = joblib.hash(est)
             check_estimator(est)
-        assert_equal(old_hash, _joblib.hash(est))
+        assert old_hash == joblib.hash(est)
 
-        with ignore_warnings(category=(FutureWarning, DeprecationWarning)):
+        with ignore_warnings(category=FutureWarning):
             # when 'est = SGDClassifier()'
             est = Estimator()
-            set_checking_parameters(est)
+            _set_checking_parameters(est)
             set_random_state(est)
             # with fitting
             est.fit(iris.data + 10, iris.target)
-            old_hash = _joblib.hash(est)
+            old_hash = joblib.hash(est)
             check_estimator(est)
-        assert_equal(old_hash, _joblib.hash(est))
+        assert old_hash == joblib.hash(est)
 
 
 def test_check_estimators_unfitted():
     # check that a ValueError/AttributeError is raised when calling predict
     # on an unfitted estimator
-    msg = "AttributeError or ValueError not raised by predict"
+    msg = "NotFittedError not raised by predict"
     assert_raises_regex(AssertionError, msg, check_estimators_unfitted,
                         "estimator", NoSparseClassifier())
 
@@ -444,11 +498,11 @@ def test_check_estimators_unfitted():
 
 
 def test_check_no_attributes_set_in_init():
-    class NonConformantEstimatorPrivateSet:
+    class NonConformantEstimatorPrivateSet(BaseEstimator):
         def __init__(self):
             self.you_should_not_set_this_ = None
 
-    class NonConformantEstimatorNoParamSet:
+    class NonConformantEstimatorNoParamSet(BaseEstimator):
         def __init__(self, you_should_set_this_=None):
             pass
 
@@ -471,7 +525,7 @@ def test_check_no_attributes_set_in_init():
 
 def test_check_estimator_pairwise():
     # check that check_estimator() works on estimator with _pairwise
-    # kernel or  metric
+    # kernel or metric
 
     # test precomputed kernel
     est = SVC(kernel='precomputed')
@@ -482,6 +536,19 @@ def test_check_estimator_pairwise():
     check_estimator(est)
 
 
+def test_check_estimator_required_parameters_skip():
+    class MyEstimator(BaseEstimator):
+        _required_parameters = ["special_parameter"]
+
+        def __init__(self, special_parameter):
+            self.special_parameter = special_parameter
+
+    assert_raises_regex(SkipTest, r"Can't instantiate estimator MyEstimator "
+                                  r"which requires parameters "
+                                  r"\['special_parameter'\]",
+                                  check_estimator, MyEstimator)
+
+
 def run_tests_without_pytest():
     """Runs the tests in this file without using pytest.
     """
